import numpy as np
import tensorflow as tf
import itertools
from matplotlib import pyplot as plt


def viz_neigbors_imgs(example,
                      example_class,
                      neighbors,
                      labels=None,
                      fig_size=(24, 4),
                      cmap='viridis'):
    """Display images nearest neighboors

    Args:
        example ([type]): [description]
        example_class ([type]): [description]
        neighbors ([type]): [description]
        labels ([type], optional): [description]. Defaults to None.
        fig_size ([type], optional): [description]. Defaults to (24, 4).
        cmap ([type], optional): [description]. Defaults to 'viridis'.
    """
    num_cols = len(neighbors) + 1
    plt.subplots(1, num_cols, figsize=fig_size)
    plt_idx = 1

    # draw target
    plt.subplot(1, num_cols, plt_idx)
    plt.imshow(example, cmap=cmap)
    plt.xticks([])
    plt.yticks([])

    val = labels[example_class] if labels else example_class
    plt.title('class %s' % val)
    plt_idx += 1

    for nbg in neighbors:
        plt.subplot(1, num_cols, plt_idx)
        val = labels[nbg.label] if labels else nbg.label
        legend = "%s - d:%.5f" % (val, nbg.distance)
        if nbg.label == example_class:
            color = cmap
        else:
            color = 'Reds'
        plt.imshow(nbg.data, cmap=color)
        plt.title(legend)
        plt.xticks([])
        plt.yticks([])

        plt_idx += 1
    plt.show()


def confusion_matrix(y_pred,
                     y_true,
                     normalize=True,
                     labels=None,
                     title='Confusion matrix',
                     cmap='Blues'):
    """Plot confusion matrix

    Args:
        y_pred: Model prediction returned by `model.match()`

        y_true: Expected class_id.

        normalize: Normalizes matrix values between 0 and 1.
        Defaults to True.

        labels: List of class string label to display instead of the class
        numerical ids. Defaults to None.

        title: Title of the confusion matrix. Defaults to 'Confusion matrix'.

        cmap: Color schema as CMAP. Defaults to 'Blues'.
    """

    cm = np.array(tf.math.confusion_matrix(y_true, y_pred))
    accuracy = np.trace(cm) / np.sum(cm).astype('float')
    misclass = 1 - accuracy

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    plt.figure(figsize=(8, 6))
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    plt.colorbar()

    if labels is not None:
        tick_marks = np.arange(len(labels))
        plt.xticks(tick_marks, labels, rotation=45)
        plt.yticks(tick_marks, labels)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    thresh = cm.max() / 1.5 if normalize else cm.max() / 2
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        val = cm[i, j]
        color = "white" if val > thresh else "black"
        txt = "%.2f" % val if val > 0.0 else "0"
        plt.text(j, i, txt, horizontalalignment="center", color=color)

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label\naccuracy={:0.4f}; misclass={:0.4f}'.format(
        accuracy, misclass))
    plt.show()



import random
from collections import defaultdict
from typing import Optional, Tuple, Sequence

import tensorflow as tf
from tensorflow_similarity.types import FloatTensor, IntTensor, Tensor
from tqdm.auto import tqdm

from .samplers import Augmenter, Sampler
from .utils import select_examples


class MultiShotMemorySampler(Sampler):
    # FIXME: typing x, y is hard.
    def __init__(self,
                 x,
                 y,
                 classes_per_batch: int = 2,
                 examples_per_class_per_batch: int = 2,
                 steps_per_epoch: int = 1000,
                 class_list: Sequence[int] = None,
                 total_examples_per_class: int = None,
                 augmenter: Optional[Augmenter] = None,
                 warmup: int = -1):
        """Create a Multishot in memory sampler that ensures that each batch is
        well balanced. That is, each batch aims to contain
        `example_per_class` examples of `class_per_batch`
        classes.

        The `batch_size` used during training will be equal to:
        `class_per_batch * example_per_class` unless an `augmenter` that alters
        the number of examples returned is used. Then the batch_size is a
        function of how many augmented examples are returned by
        the `augmenter`.


        Multishot samplers are to be used when you have multiple
        examples for the same class. If this is not the case, then see
        the [SingleShotMemorySampler()](single_memory.md) for using single
        example with augmentation.

        Memory samplers are good for datasets that fit in memory. If you have
        larger ones that needs to sample from disk then use the
        [TFRecordDatasetSampler()](tfdataset_sampler.md)


        Args:
            x: examples.

            y: labels.

            class_per_batch: Numbers of distinct class to include in a
            single batch

            examples_per_class_per_batch: How many example of each class
            to use per batch. Defaults to 2.

            steps_per_epoch: How many steps/batches per epoch.
            Defaults to 1000.

            class_list: Filter the list of examples to only keep those who
            belong to the supplied class list.

            total_examples_per_class: Restrict the number of examples for EACH
            class to total_examples_per_class if set. If not set, all the
            available examples are selected. Defaults to None - no selection.


            augmenter: A function that takes a batch in and return a batch out.
            Can alters the number of examples returned which in turn change the
            batch_size used. Defaults to None.

            warmup: Keep track of warmup epochs and let the augmenter knows
            when the warmup is over by passing along with each batch data a
            boolean `is_warmup`. See `self.get_examples()` Defaults to 0.
        """

        super().__init__(
            classes_per_batch,
            examples_per_class_per_batch=examples_per_class_per_batch,  # noqa
            steps_per_epoch=steps_per_epoch,
            augmenter=augmenter,
            warmup=warmup)

        # precompute information we need
        if not class_list:
            self.class_list = list(set([int(e) for e in y]))
        else:
            # dedup in case user mess up and cast in case its a tensor
            self.class_list = list(set([int(c) for c in class_list]))

        if classes_per_batch > len(self.class_list):
            raise ValueError("the value of class_per_batch must be <= to the\
                                number of existing class in the dataset")

        # filter
        x, y = select_examples(x,
                               y,
                               class_list=self.class_list,
                               num_examples_per_class=total_examples_per_class)

        # assign after potential selection
        self.x = x
        self.y = y

        # we need to reindex here  as the numbers of samples might have
        # changed due to the filtering
        # In this sampler, contrary to file based one, the pool of examples
        # wont' change so we can optimize by doing this step in the constructor
        self.index_per_class = defaultdict(list)
        cls = [int(c) for c in y]  # need to cast as  tensor lookup are slowww
        for idx in tqdm(range(len(x)), desc='indexing classes'):
            cl = cls[idx]
            self.index_per_class[cl].append(idx)

    def get_examples(self, batch_id: int, num_classes: int,
                     examples_per_class: int) -> Tuple[Tensor, Tensor]:

        # select class at ramdom
        class_list = random.sample(self.class_list, k=num_classes)

        # get example for each class
        idxs = []
        for class_id in class_list:
            class_idxs = self.index_per_class[class_id]
            idxs.extend(random.choices(class_idxs, k=examples_per_class))

        random.shuffle(idxs)
        batch_x = tf.gather(self.x, idxs[:self.batch_size])
        batch_y = tf.gather(self.y, idxs[:self.batch_size])

        return batch_x, batch_y


class SingleShotMemorySampler(Sampler):
    def __init__(self,
                 x: FloatTensor,
                 augmenter: Augmenter,
                 class_per_batch: int,
                 steps_per_epoch: int = 1000,
                 warmup: int = -1) -> None:
        """Create a single shot in memory sampler.

        The `batch_size` used during training will be equal to:
        `class_per_batch * example_per_class` unless the `augmenter`
        used to create similar looking examples have a different regime.
        In that case the batch_size would be a function of how many augmented
        examples are returned by the `augmenter` function.

        Single shot samplers are used when you have a single example per class
        and solely rely on data augmentation to generate similar looking
        examples.

        If you do have multiple examples per class, you should use the
        [MultiShotMemorySampler()](multishot_memory.md).

        Memory samplers are good for datasets that fit in memory. If you have
        larger ones that needs to sample from disk then use the
        [TFRecordDatasetSampler()](tfdataset_sampler.md).


        Args:

            x: Input data. The sampler assumes that each element of X is from a
            distinct class.

            augmenter: A function that takes a batch of single examples and
            return a batch out with additional examples per class.

            steps_per_epoch: How many steps/batch per epoch. Defaults to 1000.

            class_per_batch: effectively the number of element to pass to the
            augmenter for each batch request in the single shot setting.

            warmup: Keep track of warmup epochs and let the augmenter knows
            when the warmup is over by passing along with each batch data a
            boolean `is_warmup`. See `self.get_examples()` Defaults to 0.
        """

        super().__init__(class_per_batch,
                         steps_per_epoch=steps_per_epoch,
                         augmenter=augmenter,
                         warmup=warmup)
        self.x = x
        self.num_elts = len(x)

    def get_examples(self, batch_id: int, num_classes: int,
                     example_per_class: int) -> Tuple[Tensor, Tensor]:
        _ = batch_id
        _ = example_per_class

        # note: we draw at random the class so the sampler can scale up to
        # millions of points. Shuffling array is simply too slow
        idxs = tf.random.uniform((num_classes, ),
                                 minval=0,
                                 maxval=self.num_elts,
                                 dtype='int32')
        # ! don't cast data as different model use different type.
        y = tf.constant([int(i) for i in idxs])
        x = tf.constant([self.x[idx] for idx in y])

        return x, y



import tensorflow as tf
from tensorflow_similarity.samplers import select_examples
from tensorflow_similarity.samplers import MultiShotMemorySampler
import pytest


def test_valid_class_numbers():
    "Check that sampler properly detect if num_class requests >> class avail"
    y = tf.constant([1, 2, 3, 1, 2, 3, 1])
    x = tf.constant([10, 20, 30, 10, 20, 30, 10])

    class_per_batch = 42

    with pytest.raises(ValueError):
        MultiShotMemorySampler(x=x, y=y,
                               classes_per_batch=class_per_batch)


@pytest.mark.parametrize("example_per_class", [2, 20])
def test_select_examples(example_per_class):
    """Test select_examples with various sizes.

    Users may sample with replacement when creating batches, so check that we
    can handle when elements per class is either less than or greater than the
    total count of elements in the class.
    """
    y = tf.constant([1, 2, 3, 1, 2, 3, 1])
    x = tf.constant([10, 20, 30, 10, 20, 30, 10])
    cls_list = [1, 3]
    batch_x, batch_y = select_examples(x, y, cls_list, example_per_class)

    assert len(batch_y) == len(cls_list) * example_per_class
    assert len(batch_x) == len(cls_list) * example_per_class

    for x, y in zip(batch_x, batch_y):
        assert y in cls_list

        if y == 1:
            assert x == 10
        elif y == 3:
            assert x == 30


@pytest.mark.parametrize("example_per_class", [2, 20])
def test_multi_shot_memory_sampler(example_per_class):
    """Test MultiShotMemorySampler with various sizes.

    Users may sample with replacement when creating batches, so check that we
    can handle when elements per class is either less than or greater than the
    total count of elements in the class.
    """
    y = tf.constant([1, 2, 3, 1, 2, 3, 1])
    x = tf.constant([10, 20, 30, 10, 20, 30, 10])
    class_per_batch = 2
    batch_size = example_per_class * class_per_batch

    ms_sampler = MultiShotMemorySampler(x=x,
                                        y=y,
                                        classes_per_batch=class_per_batch,
                                        examples_per_class_per_batch=example_per_class)  # noqa

    batch_x, batch_y = ms_sampler.generate_batch(batch_id=606)

    assert len(batch_y) == batch_size
    assert len(batch_x) == batch_size
    num_classes, _ = tf.unique(batch_y)
    assert len(num_classes) == class_per_batch

    for x, y in zip(batch_x, batch_y):
        if y == 1:
            assert x == 10
        elif y == 2:
            assert x == 20
        elif y == 3:
            assert x == 30



import pytest
import tensorflow as tf
import numpy as np
from tensorflow_similarity.distances import CosineDistance, InnerProductSimilarity
from tensorflow_similarity.distances import EuclideanDistance
from tensorflow_similarity.distances import ManhattanDistance
from tensorflow_similarity.distances import distance_canonicalizer
from tensorflow_similarity.distances import DISTANCES


def test_distance_mapping():
    for d in DISTANCES:

        # self naming
        d2 = distance_canonicalizer(d.name)
        assert d2.name == d.name

        # aliases
        for a in d.aliases:
            d2 = distance_canonicalizer(a)
            assert d2.name == d.name


def test_distance_passthrough():
    "Canonilizer is expected to return distance object as is"
    d = EuclideanDistance()
    d2 = distance_canonicalizer(d)
    assert d == d2


def test_non_existing_distance():
    with pytest.raises(ValueError):
        distance_canonicalizer('notadistance')


def angular_distance_np(feature):
    """Computes the angular distance matrix in numpy.
    Args:
      feature: 2-D numpy array of size [number of data, feature dimension]
    Returns:
      angular_distances: 2-D numpy array of size
        [number of data, number of data].
    """

    # l2-normalize all features
    normed = feature / np.linalg.norm(feature, ord=2, axis=1, keepdims=True)
    cosine_similarity = normed @ normed.T
    inverse_cos_sim = 1 - cosine_similarity

    return inverse_cos_sim


def test_cosine_same():
    # pairwise
    a = tf.convert_to_tensor([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0]])
    d = CosineDistance()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 0


def test_cosine_opposite():
    a = tf.convert_to_tensor([[0.0, 1.0], [1.0, 0.0]])
    d = CosineDistance()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 2


def test_cosine_vals():
    a = tf.constant([[0.1, 0.3, 0.2], [0.0, 0.1, 0.5]])
    d = CosineDistance()
    vals = d(a)
    assert vals[0][0] == 0
    assert vals[0][1] == 0.31861484


def test_euclidean():
    a = tf.convert_to_tensor([[0.0, 3.0], [4.0, 0.0]])
    d = EuclideanDistance()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 10


def test_euclidean_same():
    a = tf.convert_to_tensor([[1.0, 1.0], [1.0, 1.0]])
    d = EuclideanDistance()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 0


def test_euclidean_opposite():
    a = tf.convert_to_tensor([[0.0, 1.0], [0.0, -1.0]])
    d = EuclideanDistance()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 4


def test_manhattan():
    a = tf.convert_to_tensor([
        [0.0, 0.0],
        [0.0, 1.0],
        [1.0, 1.0],
        [3.0, 0.0]
    ])
    d = ManhattanDistance()
    vals = d(a)
    expected = tf.constant([
        [0.0, 1.0, 2.0, 3.0],
        [1.0, 0.0, 1.0, 4.0],
        [2.0, 1.0, 0.0, 3.0],
        [3.0, 4.0, 3.0, 0.0]
    ])
    assert tf.reduce_all(tf.math.equal(vals, expected))


def test_manhattan_same():
    a = tf.convert_to_tensor([[1.0, 1.0], [1.0, 1.0]])
    d = ManhattanDistance()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 0


def test_manhattan_opposite():
    a = tf.convert_to_tensor([[0.0, 1.0], [0.0, -1.0]])
    d = ManhattanDistance()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 4


def test_innerprod():
    a = [[1, 2, 3], [1, 3, 3]]
    d = InnerProductSimilarity()
    vals = d(a)
    assert tf.round(tf.reduce_sum(vals)) == 65



import tensorflow as tf
from tensorflow_similarity.distance_metrics import DistanceMetric
from tensorflow_similarity.distance_metrics import DistanceGapMetric
from tensorflow_similarity.distances import CosineDistance
from tensorflow_similarity.types import FloatTensor

EMB1 = [
    [0.5, 1, 0.5],
    [0.2, 0.8, 0.4],
]
EMB2 = [
    [0.66, 0.5, 0.34],
    [0.77, 0.9, 0.92],
]

LABELS = tf.Variable([[1], [1], [2], [2]], dtype='int32')
EMBEDDINGS = tf.Variable(EMB1 + EMB2)


def cosine(a: FloatTensor, b: FloatTensor, axis: int = -1) -> FloatTensor:
    t1 = tf.nn.l2_normalize(a, axis=axis)
    t2 = tf.nn.l2_normalize(b, axis=axis)
    distances = 1 - tf.linalg.matmul(t1, t2, transpose_b=True)
    distances = tf.math.maximum(distances, 0.0)
    return distances


def compute_metric(distance, aggregate, labels, embeddings):
    'Inner function that call the core class'
    metric = DistanceMetric(distance, aggregate=aggregate)
    metric.update_state(labels, embeddings, None)
    return metric.result()


def manual_hard_mining(mydistance, E1, E2):
    l1 = tf.reduce_max(mydistance(E1, E1), axis=1)
    l2 = tf.reduce_max(mydistance(E2, E2), axis=1)
    return [l1, l2]


def test_distance_metric_serialize():
    metric = DistanceMetric('cosine', aggregate='max')
    config = metric.get_config()
    metric2 = DistanceMetric.from_config(config)
    assert isinstance(metric2.distance, CosineDistance)
    assert metric2.aggregate == metric.aggregate

    metric.update_state(LABELS, EMBEDDINGS, None)
    metric2.update_state(LABELS, EMBEDDINGS, None)
    assert metric.result() == metric2.result()


class DistanceMetricsTest(tf.test.TestCase):

    def test_avg_positive(self):

        agg = ['avg', tf.reduce_mean]

        # metric computation
        metric_val = compute_metric('cosine', agg[0], LABELS, EMBEDDINGS)

        # manual computation
        hard_distances = manual_hard_mining(cosine, EMB1, EMB2)
        manual_val = agg[1](hard_distances)
        self.assertAllClose(metric_val, manual_val)

    def test_sum_positive(self):
        agg = ['sum', tf.reduce_sum]

        # metric computation
        metric_val = compute_metric('cosine', agg[0], LABELS, EMBEDDINGS)

        # manual computation
        hard_distances = manual_hard_mining(cosine, EMB1, EMB2)
        manual_val = agg[1](hard_distances)
        self.assertAllClose(metric_val, manual_val)

    def test_max_positive(self):
        agg = ['max', tf.reduce_max]

        # metric computation
        metric_val = compute_metric('cosine', agg[0], LABELS, EMBEDDINGS)

        # manual computation
        hard_distances = manual_hard_mining(cosine, EMB1, EMB2)
        manual_val = agg[1](hard_distances)
        self.assertAllClose(metric_val, manual_val)

    def test_min_positive(self):
        agg = ['min', tf.reduce_min]

        # metric computation
        metric_val = compute_metric('cosine', agg[0], LABELS, EMBEDDINGS)

        # manual computation
        hard_distances = manual_hard_mining(cosine, EMB1, EMB2)
        manual_val = agg[1](hard_distances)
        self.assertAllClose(metric_val, manual_val)


def test_gap():
    metric = DistanceGapMetric('cosine')
    metric.update_state(LABELS, EMBEDDINGS, None)
    metric.result()



import math
from copy import copy
from tqdm.auto import tqdm
from .evaluator import Evaluator
from collections import defaultdict
import tensorflow as tf
from typing import DefaultDict, List, Dict, Union
from tensorflow_similarity.metrics import EvalMetric, make_metric
from tensorflow_similarity.types import Lookup


class MemoryEvaluator(Evaluator):
    """In memory index performance evaluation and calibration."""

    def evaluate(self,
                 index_size: int,
                 metrics: List[Union[str, EvalMetric]],
                 targets_labels: List[int],
                 lookups: List[List[Lookup]],
                 distance_rounding: int = 8
                 ) -> Dict[str, Union[float, int]]:
        """Evaluates lookup performances against a supplied set of metrics

        Args:
            index_size: Size of the search index.

            metrics: List of `EvalMetric()` to evaluate lookup matches against.

            targets_labels: List of the expected labels to match.

            lookups: List of lookup results as produced by the
            `Index().batch_lookup()` method.

            distance_rounding: How many digit to consider to decide if
            the distance changed. Defaults to 8.

        Returns:
            Dictionnary of metric results where keys are the metric
            names and values are the metrics values.
        """
        # [nn[{'distance': xxx}, ]]
        # normalize metrics
        eval_metrics: List[EvalMetric] = [make_metric(m) for m in metrics]

        # get max_k from first lookup result
        max_k = len(lookups[0])

        # ! don't add intermediate computation that don't speedup multiple
        # ! metrics. Those goes into the metrics themselves.
        # compute intermediate representations used by metrics
        # rank 0 == no match / distance 0 == unknown
        num_matched = 0
        match_ranks = [0] * len(targets_labels)
        match_distances = [0.0] * len(targets_labels)
        for lidx, lookup in enumerate(lookups):
            true_label = targets_labels[lidx]
            for nidx, n in enumerate(lookup):
                rank = nidx + 1
                if n.label == true_label:
                    # print(n['label'], true_label, lookup)
                    match_ranks[lidx] = rank
                    match_distances[lidx] = (
                            round(n.distance, distance_rounding))
                    num_matched += 1

        # compute metrics
        evaluation = {}
        for m in eval_metrics:
            evaluation[m.name] = m.compute(
                max_k,
                targets_labels,
                index_size,
                match_ranks,
                match_distances,
                lookups,  # e.g used when k > 1 for confusion matrix
            )

        return evaluation

    def calibrate(self,
                  index_size: int,
                  calibration_metric: EvalMetric,
                  thresholds_targets: Dict[str, float],
                  targets_labels: List[int],
                  lookups: List[List[Lookup]],
                  extra_metrics: List[Union[str, EvalMetric]] = [],
                  distance_rounding: int = 8,
                  metric_rounding: int = 6,
                  verbose: int = 1):
        """Computes the distances thresholds that the calibration much match to
        meet fixed target.

        Args:
            index_size: Index size.

            calibration_metric: Metric used for calibration.

            thresholds_targets: Calibration metrics thresholds that are
            targeted. The function will find the closed distance value.

            targets_labels: List of expected labels for the lookups.

            lookup: List of lookup results as produced by the
            `Index.batch_lookup()` method.

            extra_metrics: Additional metrics that should be computed and
            reported as part of the calibration. Defaults to [].

            distance_rounding: How many digit to consider to
            decide if the distance changed. Defaults to 8.

            metric_rounding: How many digit to consider to decide if
            the metric changed. Defaults to 6.

            verbose: Be verbose. Defaults to 1.
        """

        # distance are rounded because of numerical instablity
        # copy threshold targets as we are going to delete them and don't want
        # to alter users supplied data
        thresholds_targets = copy(thresholds_targets)

        # making a single list of metrics
        # Need expl covariance problem
        combined_metrics = list(extra_metrics)
        combined_metrics.append(calibration_metric)

        # data preparation: flatten and rounding
        distances = []
        for lu in lookups:
            for n in lu:
                distances.append(round(n.distance, distance_rounding))

        targets_labels = [int(i) for i in targets_labels]

        # sorting them
        # !keep the casting to int() or it will be awefully slow
        sorted_distances_idxs = [int(i) for i in list(tf.argsort(distances))]
        sorted_distances_values = [distances[i] for i in sorted_distances_idxs]
        num_distances = len(distances)

        # evaluating performance as distance value increase
        evaluations = []
        if verbose:
            pb = tqdm(total=num_distances, desc='Evaluating')

        for dist in sorted_distances_values:
            # update distance theshold for metrics
            for m in combined_metrics:
                if isinstance(m, EvalMetric):  # typechecking requires this
                    m.distance_threshold = dist

            res = self.evaluate(index_size, combined_metrics, targets_labels,
                                lookups, distance_rounding)
            res['distance'] = float(dist)  # ! cast needed for serialization
            evaluations.append(res)
            if verbose:
                pb.update()

        if verbose:
            pb.close()

        # find the thresholds by going from right to left

        # which direction metric improvement is?
        # !loop is right to left so max is decreasing and min is increasing
        if calibration_metric.direction == 'max':
            # we want the lowest value at the largest distance possible
            cmp = self._is_lower
            prev_value = math.inf  # python 3.x only
        else:
            # we want the highest value at the largest distance possible
            cmp = self._is_higher
            prev_value = 0

        # we need a collection of list to apply vectorize operations and make
        # the analysis / viz of the calibration data signifcantly easier
        thresholds: DefaultDict[str, List[Union[int, float]]] = defaultdict(list)  # noqa
        cutpoints: DefaultDict[str, Dict[str, Union[str, float, int]]] = defaultdict(dict)  # noqa
        num_distances = len(sorted_distances_values)
        if verbose:
            pb = tqdm(total=num_distances, desc='computing thresholds')

        # looping from right to left as we want the max distance for a given
        # metric value
        for ridx in range(num_distances):
            idx = num_distances - ridx - 1  # reversed

            # Rounding the calibration metric to create bins
            curr_eval = evaluations[idx]
            calibration_value = curr_eval[calibration_metric.name]
            curr_value = round(calibration_value, metric_rounding)

            # ? if bug use this line check that the values evolve correclty.
            # print(curr_value, prev_value, cmp(curr_value, prev_value))

            if cmp(curr_value, prev_value):

                # add a new distance threshold
                thresholds['value'].append(curr_value)

                # ! the correct distance is already in the eval data
                # record the value for all the metrics requested by the user
                for k, v in curr_eval.items():
                    thresholds[k].append(v)

                # update current threshold value
                prev_value = curr_value

                # check if the current value meet or exceed threshold target
                to_delete = []  # can't delete in an interation loop
                for name, value in thresholds_targets.items():
                    if cmp(curr_value, value, equal=True):
                        cutpoints[name] = {'name': name}  # useful for display
                        for k in thresholds.keys():
                            cutpoints[name][k] = thresholds[k][-1]
                        to_delete.append(name)

                # removing found targets to avoid finding lower value
                # recall we go from right to left in the evaluation
                for name in to_delete:
                    del thresholds_targets[name]

            if verbose:
                pb.update()

        if verbose:
            pb.close()

        # find the optimal cutpoint
        if calibration_metric.direction == 'min':
            best_idx = tf.math.argmin(thresholds[calibration_metric.name])
        else:
            best_idx = tf.math.argmax(thresholds[calibration_metric.name])

        # record its value
        cutpoints['optimal'] = {'name': 'optimal'}  # useful for display
        for k in thresholds.keys():
            cutpoints['optimal'][k] = thresholds[k][best_idx]

        # reverse the threshold so they go from left to right as user expect
        for k in thresholds.keys():  # this syntax is need for mypy ...
            thresholds[k].reverse()

        return thresholds, cutpoints

    def _is_lower(self, curr, prev, equal=False):
        if equal:
            return curr <= prev
        return curr < prev

    def _is_higher(self, curr, prev, equal=False):
        if equal:
            return curr >= prev
        return curr > prev



import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras import layers


@tf.keras.utils.register_keras_serializable(package="Similarity")
class MetricEmbedding(Layer):
    def __init__(self, size):
        """Normalized embedding layer
        Args:
            size (int): size of the embbeding. Usually something like 32, 64 or
            256 floats.
        """
        self.size = size
        self.dense = layers.Dense(size)
        super(MetricEmbedding, self).__init__()
        # FIXME: enforce the shape
        # self.input_spec = rank2

    def call(self, inputs):
        x = self.dense(inputs)
        return tf.math.l2_normalize(x, axis=1)

    def get_config(self):
        return {
            'size': self.size
        }



from abc import ABC, abstractmethod
from typing import List, Dict, Union
from tensorflow_similarity.metrics import EvalMetric
from tensorflow_similarity.types import Lookup


class Evaluator(ABC):
    """Evaluates index performance and calibrates it.

    Index evaluators are derived from this abstract class to allow users to
    override the evaluation module to use additional data or interface
    with existing evaluation system. For example allowing to fetch data from
    a remote database.
    """

    @abstractmethod
    def evaluate(self,
                 index_size: int,
                 metrics: List[Union[str, EvalMetric]],
                 targets_labels: List[int],
                 lookups: List[List[Lookup]],
                 distance_rounding: int = 8
                 ) -> Dict[str, Union[float, int]]:
        """Evaluates lookup performances against a supplied set of metrics

        Args:
            index_size: Size of the search index.

            metrics: List of `EvalMetric()` to evaluate lookup matches against.

            targets_labels: List of the expected labels to match.

            lookups: List of lookup results as produced by the
            `Index().batch_lookup()` method.

            distance_rounding: How many digit to consider to decide if
            the distance changed. Defaults to 8.

        Returns:
            Dictionnary of metric results where keys are the metric
            names and values are the metrics values.
        """

    @abstractmethod
    def calibrate(self,
                  index_size: int,
                  calibration_metric: EvalMetric,
                  thresholds_targets: Dict[str, float],
                  targets_labels: List[int],
                  lookups: List[List[Lookup]],
                  extra_metrics: List[Union[str, EvalMetric]] = [],
                  distance_rounding: int = 8,
                  metric_rounding: int = 6,
                  verbose: int = 1):
        """Computes the distances thresholds that the calibration much match to
        meet fixed target.

        Args:
            index_size: Index size.

            calibration_metric: Metric used for calibration.

            thresholds_targets: Calibration metrics thresholds that are
            targeted. The function will find the closed distance value.

            targets_labels: List of expected labels for the lookups.

            lookup: List of lookup results as produced by the
            `Index.batch_lookup()` method.

            extra_metrics: Additional metrics that should be computed and
            reported as part of the calibration. Defaults to [].

            distance_rounding: How many digit to consider to
            decide if the distance changed. Defaults to 8.

            metric_rounding: How many digit to consider to decide if
            the metric changed. Defaults to 6.

            verbose: Be verbose. Defaults to 1.
        """



from tqdm.auto import tqdm
from collections import defaultdict
import random
from typing import Sequence, Tuple

import tensorflow as tf
from tensorflow_similarity.types import IntTensor, FloatTensor


def select_examples(x: FloatTensor,
                    y: IntTensor,
                    class_list: Sequence[int] = None,
                    num_examples_per_class: int = None,
                    ) -> Tuple[FloatTensor, IntTensor]:
    """Randomly select at most N examples per class

    Args:
        x: A 2-D Tensor containing the data.

        y: A 1-D Tensor containing the labels.

        class_list: Filter the list of examples to only keep thoses those who
        belong to the supplied class list. In no class is supplied, keep
        examples for all the classes. Default to None - keep all the examples.

        num_examples_per_class: Restrict the number of examples for EACH
        class to num_examples_per_class if set. If not set, all the available
        examples are selected. Defaults to None - no selection.

    Returns:
        A Tuple containing the subset of x and y.
    """

    # cast class_list if it exist to avoid slowness
    if class_list is not None:
        class_list_int = [int(c) for c in class_list]
    else:
        class_list_int = list(set([int(e) for e in y]))

    # Mapping class to idx
    index_per_class = defaultdict(list)
    cls = [int(c) for c in y]

    for idx in tqdm(range(len(x)), desc="filtering classes"):
        cl = cls[idx]  # need to cast tensor

        # if user provided a class_list, check it's part of it.
        if class_list is not None and cl in class_list_int:
            index_per_class[cl].append(idx)
        else:
            # just add all class examples
            index_per_class[cl].append(idx)

    # restrict numbers of samples
    idxs = []
    for class_id in tqdm(class_list_int, desc="selecting examples"):
        class_idxs = index_per_class[class_id]

        # restrict num examples?
        if num_examples_per_class:
            idxs.extend(random.choices(class_idxs, k=num_examples_per_class))
        else:
            idxs.extend(class_idxs)

    random.shuffle(idxs)
    batch_x = tf.gather(x, idxs)
    batch_y = tf.gather(y, idxs)

    return batch_x, batch_y



from .utils import select_examples  # noqa
from .memory_samplers import MultiShotMemorySampler  # noqa
from .memory_samplers import SingleShotMemorySampler  # noqa
from .tfrecords_samplers import TFRecordDatasetSampler  # noqa
from .tfdataset_samplers import TFDatasetMultiShotMemorySampler  # noqa



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Core TensorFlow types."""

import dataclasses
from typing import Any, Callable, Iterable, Union, Optional, List, Tuple, Optional  # noqa

import numpy as np
import tensorflow as tf


class PandasDataFrame(object):
    """Symbolic pandas frame
    Pandas type are too loose you get an Any. We want a PandaFrame
    """
    pass


class Tensor(List):
    """The base class of all dense Tensor objects.

    A dense tensor has a static data type (dtype), and may have a static rank
    and shape. Tensor objects are immutable. Mutable objects may be backed by
    a Tensor which holds the unique handle that identifies the mutable object.
    """
    @property
    def dtype(self):
        pass

    @property
    def shape(self):
        pass

    @property
    def __len__(self):
        pass

    @property
    def __iter__(self):
        pass


class BoolTensor(Tensor):
    """Bool tensor """


class FloatTensor(Tensor):
    """Float tensor """


class IntTensor(Tensor):
    """Integer tensor"""


class Symbol(Tensor):
    """Symbolic "graph" Tensor.

    These objects represent the output of an op definition and do not carry a
    value.
    """
    pass


class Value(Tensor):
    """Tensor that can be associated with a value (aka "eager tensor").

    These objects represent the (usually future) output of executing an op
    immediately.
    """
    def numpy(self):
        pass


EqFunc = Callable[[Any, Any], bool]


def _optional_eq(a: Any, b: Any, eq_fun: EqFunc) -> bool:
    """__eq__ for Optional[Any] types."""
    if a is None:
        return b is None
    elif b is None:
        return False

    return eq_fun(a, b)


def _basic_eq(a: Any, b: Any) -> bool:
    eq: bool = a == b
    return eq


def _ndarray_eq(a: np.ndarray, b: np.ndarray) -> bool:
    eq: bool = np.allclose(a, b, rtol=0, atol=0, equal_nan=True)
    return eq


def _tf_eq(a: Tensor, b: Tensor) -> bool:
    eq: bool = tf.math.reduce_all(tf.math.equal(a, b))
    return eq


@dataclasses.dataclass
class Lookup:
    """Metadata associated with a query match.

    Attributes:
        rank: Rank of the match with respect to the query distance.

        distance: The distance from the match to the query.

        label: The label associated with the match. Default None.

        embedding: The embedded match vector. Default None.

        data: The original Tensor representation of the match result.
        Default None.
    """
    rank: int
    distance: float
    label: Optional[int] = dataclasses.field(default=None)
    embedding: Optional[np.ndarray] = dataclasses.field(default=None)
    data: Optional[Tensor] = dataclasses.field(default=None)

    def __eq__(self, other) -> bool:
        if other.__class__ is not self.__class__:
            return False
        if self.rank != other.rank:
            return False
        if self.distance != other.distance:
            return False
        if not _optional_eq(self.label, other.label, _basic_eq):
            return False
        if not _optional_eq(self.embedding, other.embedding, _ndarray_eq):
            return False
        if not _optional_eq(self.data, other.data, _tf_eq):
            return False

        return True



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Multi Similarity Loss"""

import tensorflow as tf
from typing import Any, Callable, Union

from tensorflow_similarity.distances import Distance, distance_canonicalizer
from tensorflow_similarity.algebra import build_masks, masked_max, masked_min
from tensorflow_similarity.types import FloatTensor, IntTensor
from .metric_loss import MetricLoss
from .utils import logsumexp


@tf.keras.utils.register_keras_serializable(package="Similarity")
@tf.function
def multisimilarity_loss(labels: IntTensor,
                         embeddings: FloatTensor,
                         distance: Callable,
                         alpha: float = 2.0,
                         beta: float = 40,
                         epsilon: float = 0.2,
                         lmda: float = 1.0) -> Any:
    """Multi Similarity loss computations

    Args:
        labels: labels associated with the embed.

        embeddings: Embedded examples.

        distance: Which distance function to use to compute the pairwise.

        alpha: The exponential weight for the positive pairs. Increasing alpha
        makes the logsumexp softmax closer to the max positive pair distance,
        while decreasing it makes it closer to max(P) + log(batch_size).

        beta: The exponential weight for the negative pairs. Increasing beta
        makes the logsumexp softmax closer to the max negative pair distance,
        while decreasing it makes the softmax closer to
        max(N) + log(batch_size).

        epsilon: Used to remove easy positive and negative pairs. We only keep
        positives that we greater than the (smallest negative pair - epsilon)
        and we only keep negatives that are less than the
        (largest positive pair + epsilon).

        lmda: Used to weight the distance. Below this distance, negatives are
        up weighted and positives are down weighted. Similarly, above this
        distance negatives are down weighted and positive are up weighted.

    Returns:
        Loss: The loss value for the current batch.
    """
    # [Label]
    # ! Weirdness to be investigated
    # do not remove this code. It is actually needed for specific situation
    # Reshape label tensor to [batch_size, 1] if not already in that format.
    # labels = tf.reshape(labels, (labels.shape[0], 1))
    batch_size = tf.size(labels)

    # [distances]
    pairwise_distances = distance(embeddings)

    # [masks]
    positive_mask, negative_mask = build_masks(labels, batch_size)

    # [pair mining using Similarity-P]
    # This is essentially hard mining the negative and positive pairs.

    # Keep all positives > Min(neg_dist - epsilon).
    neg_min, _ = masked_min(pairwise_distances, negative_mask)
    neg_min = tf.math.subtract(neg_min, epsilon)
    pos_sim_p_mask = tf.math.greater(pairwise_distances, neg_min)
    pos_sim_p_mask = tf.math.logical_and(pos_sim_p_mask, positive_mask)

    # Keep all negatives < Max(pos_dist + epsilon).
    pos_max, _ = masked_max(pairwise_distances, positive_mask)
    pos_max = tf.math.add(pos_max, epsilon)
    neg_sim_p_mask = tf.math.less(pairwise_distances, pos_max)
    neg_sim_p_mask = tf.math.logical_and(neg_sim_p_mask, negative_mask)

    # Mark all pairs where we have both valid negative and positive pairs.
    valid_anchors = tf.math.logical_and(
            tf.math.reduce_any(pos_sim_p_mask, axis=1),
            tf.math.reduce_any(neg_sim_p_mask, axis=1)
    )

    # Cast masks as floats to support multiply
    valid_anchors = tf.cast(valid_anchors, dtype='float32')
    pos_sim_p_mask_f32 = tf.cast(pos_sim_p_mask, dtype='float32')
    neg_sim_p_mask_f32 = tf.cast(neg_sim_p_mask, dtype='float32')

    # [Weight the remaining pairs using Similarity-S and Similarity-N]
    shifted_distances = pairwise_distances - lmda
    pos_dists = alpha * shifted_distances
    neg_dists = -1 * beta * shifted_distances

    # [compute loss]

    # Positive pairs with a distance above 0 will be up weighted.
    p_loss = logsumexp(pos_dists, pos_sim_p_mask_f32)
    p_loss = p_loss / alpha

    # Negative pairs with a distance below 0 will be up weighted.
    n_loss = logsumexp(neg_dists, neg_sim_p_mask_f32)
    n_loss = n_loss / beta

    # Remove any anchors that have empty neg or pos pairs.
    # NOTE: reshape is required here because valid_anchors is [m] and
    #       p_loss + n_loss is [m, 1].
    multisim_loss = tf.math.multiply(
            p_loss + n_loss,
            tf.reshape(valid_anchors, (-1, 1))
    )
    # reduce and scale loss so it isn't a function of the batch size.
    multisim_loss = tf.math.reduce_mean(multisim_loss)

    return multisim_loss


@tf.keras.utils.register_keras_serializable(package="Similarity")
class MultiSimilarityLoss(MetricLoss):
    """Computes the multi similarity loss in an online fashion.


    `y_true` must be  a 1-D integer `Tensor` of shape (batch_size,).
    It's values represent the classes associated with the examples as
    **integer  values**.

    `y_pred` must be 2-D float `Tensor`  of L2 normalized embedding vectors.
    you can use the layer `tensorflow_similarity.layers.L2Embedding()` as the
    last layer of your model to ensure your model output is properly
    normalized.
    """

    def __init__(self,
                 distance: Union[Distance, str] = 'cosine',
                 alpha: float = 1.0,
                 beta: float = 20,
                 epsilon: float = 0.2,
                 lmda: float = 0.5,
                 name: str = None):
        """Initializes the Multi Similarity Loss

        Args:
            distance: Which distance function to use to compute the pairwise
            distances between embeddings. Defaults to 'cosine'.

            alpha: The exponential weight for the positive pairs. Increasing
            alpha makes the logsumexp softmax closer to the max positive pair
            distance, while decreasing it makes it closer to
            max(P) + log(batch_size).

            beta: The exponential weight for the negative pairs. Increasing
            beta makes the logsumexp softmax closer to the max negative pair
            distance, while decreasing it makes the softmax closer to
            max(N) + log(batch_size).

            epsilon: Used to remove easy positive and negative pairs. We only
            keep positives that we greater than the (smallest negative pair -
            epsilon) and we only keep negatives that are less than the
            (largest positive pair + epsilon).

            lmda: Used to weight the distance. Below this distance, negatives
            are up weighted and positives are down weighted. Similarly, above
            this distance negatives are down weighted and positive are up
            weighted.

            name: Loss name. Defaults to None.
        """

        # distance canonicalization
        distance = distance_canonicalizer(distance)
        self.distance = distance

        super().__init__(multisimilarity_loss,
                         name=name,
                         reduction=tf.keras.losses.Reduction.NONE,
                         distance=distance,
                         alpha=alpha,
                         beta=beta,
                         epsilon=epsilon,
                         lmda=lmda)



from collections import defaultdict
from copy import copy
from pathlib import Path
from tensorflow_similarity.evaluators.evaluator import Evaluator
from typing import Dict, List, Union, DefaultDict, Optional
from tabulate import tabulate
from tqdm.auto import tqdm

import tensorflow as tf
from tensorflow.keras.optimizers import Optimizer
from tensorflow.keras.metrics import Metric
from tensorflow.keras.losses import Loss

from tensorflow_similarity.indexer import Indexer
from tensorflow_similarity.tables import Table
from tensorflow_similarity.matchers import Matcher
from tensorflow_similarity.metrics import EvalMetric, make_metric
from tensorflow_similarity.distances import Distance
from tensorflow_similarity.distances import distance_canonicalizer
from tensorflow_similarity.distance_metrics import DistanceMetric
from tensorflow_similarity.losses import MetricLoss
from tensorflow_similarity.types import FloatTensor, Lookup, IntTensor, Tensor
from tensorflow_similarity.types import PandasDataFrame


@tf.keras.utils.register_keras_serializable(package="Similarity")
class SimilarityModel(tf.keras.Model):
    """Specialized Keras.Model which implement the core features needed for
    metric learning. In particular, `SimilarityModel()` supports indexing,
    searching and saving the embeddings predicted by the network.

    All Similarity models classes derive from this class to benefits from those
    core features.
    """

    # @property
    # def _index(self):
    #     if not hasattr(self, '_index'):
    #         ValueError("Index doesn't exist: index data before quering it")
    #     return self._index

    # @index.setter
    # def _index(self, index):
    #     self._index: Indexer = index

    def compile(self,
                optimizer: Union[Optimizer, str, Dict, List] = 'rmsprop',  # noqa
                distance: Union[Distance, str] = 'auto',
                loss: Union[Loss, MetricLoss, str, Dict, List] = None,
                metrics: Union[Metric, DistanceMetric, str, Dict, List] = None,
                embedding_output: int = None,
                table: Union[Table, str] = 'memory',
                matcher: Union[Matcher, str] = 'nmslib',
                evaluator: Union[Evaluator, str] = 'memory',
                stat_buffer_size: int = 1000,
                loss_weights: List = None,
                weighted_metrics: List = None,
                run_eagerly: bool = False,
                steps_per_execution: int = 1,
                **kwargs):
        """Configures the model for training.

        Args:

            optimizer: String (name of optimizer) or optimizer instance. See
            [tf.keras.optimizers](https://www.tensorflow.org/api_docs/python/tf/keras/optimizers).

            loss: String (name of objective function), objective function,
            any `tensorflow_similarity.loss.*` instance or a
            `tf.keras.losses.Loss` instance. See the [Losses
            documentation](../losses.md) for a list of metric learning
            specifics loss offered by TensorFlow Similairy and
            [tf.keras.losses](https://www.tensorflow.org/api_docs/python/tf/keras/losses)
            for the losses available directly in TensorFlow.


            metrics: List of metrics to be evaluated by the model during
            training and testing. Each of those can be a string,
            a function or a [tensorflow_similairty.metrics.*](../metrics.md)
            instance. Note that the metrics used for some type of
            metric-learning such as distance learning (e.g via triplet loss)
            have a different prototype than the metrics used in
            standard models and you can't use the `tf.keras.metrics` for those
            type of learning.

            Additionally many distance metrics are computed based of the
            [Indexer()](../indexer.md) performance. E.g Matching Top 1
            accuracy. For technical and performance reasons, indexing data at
            each training batch to compute those is impractical so
            those metrics are computed at epoch end via
            the [EvalCallback](../callbacks.md)

            See [Evaluation Metrics](../eval_metrics.md) for a list of
            available metrics.

            For multi-output models you can specify different metrics for
            different outputs by passing a dictionary, such as
            `metrics={'similarity': 'min_neg_gap', 'other': ['accuracy', 'mse']}`.
            You can also pass a list (len = len(outputs)) of lists of metrics
            such as `metrics=[['min_neg_gap'], ['accuracy', 'mse']]` or
            `metrics=['min_neg_gap', ['accuracy', 'mse']]`. For outputs which
            are not related to metrics learning, you can use any of the
            standard `tf.keras.metrics`.


            loss_weights: Optional list or dictionary specifying scalar
            coefficients (Python floats) to weight the loss contributions of
            different model outputs. The loss value that will be minimized
            by the model will then be the *weighted sum* of all individual
            losses, weighted by the `loss_weights` coefficients.
            If a list, it is expected to have a 1:1 mapping to the model's
            outputs. If a dict, it is expected to map output names (strings)
            to scalar coefficients.

            weighted_metrics: List of metrics to be evaluated and weighted by
            sample_weight or class_weight during training and testing.


            run_eagerly: Bool. Defaults to `False`. If `True`, this `Model`'s
            logic will not be wrapped in a `tf.function`. Recommended to leave
            this as `None` unless your `Model` cannot be run inside a
            `tf.function`.

            steps_per_execution: Int. Defaults to 1. The number of batches to
            run during each `tf.function` call. Running multiple batches
            inside a single `tf.function` call can greatly improve performance
            on TPUs or small models with a large Python overhead.
            At most, one full epoch will be run each execution. If a number
            larger than the size of the epoch is passed,  the execution will be
            truncated to the size of the epoch.
            Note that if `steps_per_execution` is set to `N`,
            `Callback.on_batch_begin` and `Callback.on_batch_end` methods will
            only be called every `N` batches (i.e. before/after each
            `tf.function` execution).

    Raises:
        ValueError: In case of invalid arguments for
            `optimizer`, `loss` or `metrics`.
    """

        # Fetching the distance used from the first loss if auto
        if distance == 'auto':
            if isinstance(loss, list):
                metric_loss = loss[0]
            else:
                metric_loss = loss

            try:
                distance = metric_loss.distance
            except:  # noqa
                msg = "distance='auto' only works if the first loss is a\
                       metric loss"

                raise ValueError(msg)
            print("Distance metric automatically set to", distance,
                  "use the distance arg to override.")
        else:
            distance = distance_canonicalizer(distance)

        # check if we we need to set the embedding head
        num_outputs = len(self.output_names)
        if embedding_output and embedding_output > num_outputs:
            raise Exception("Embedding_output value exceed number of model "
                            "outputs")

        if not embedding_output and num_outputs > 1:
            print("Embedding output set to be model output 0",
                  "Use the embedding_output arg to override this")
            embedding_output = 0

        # fetch embedding size as some ANN libs requires it for init
        if num_outputs > 1:
            self.embedding_size = self.outputs[embedding_output].shape[1]
        else:
            self.embedding_size = self.outputs[0].shape[1]

        # init index
        self._index = Indexer(distance=distance,
                              embedding_size=self.embedding_size,
                              embedding_output=embedding_output,
                              table=table,
                              matcher=matcher,
                              evaluator=evaluator,
                              stat_buffer_size=stat_buffer_size)

        # call underlying keras method
        super().compile(optimizer=optimizer,
                        loss=loss,
                        metrics=metrics,
                        loss_weights=loss_weights,
                        weighted_metrics=weighted_metrics,
                        run_eagerly=run_eagerly,
                        steps_per_execution=steps_per_execution,
                        **kwargs)

    def index(self,
              x: Tensor,
              y: IntTensor = None,
              data: Optional[Tensor] = None,
              build: bool = True,
              verbose: int = 1):
        """Index data.

        Args:
            x: Samples to index.

            y: class ids associated with the data if any. Defaults to None.
            store_data: store the data associated with the samples in Table.
            Defaults to True.

            build: Rebuild the index after indexing. This is needed to make the
            new samples searchable. Set it to false to save processing time
            when calling indexing repeatidly without the need to search between
            the indexing requests. Defaults to True.

            verbose: Output indexing progress info. Defaults to 1.
        """

        if not self._index:
            raise Exception('You need to compile the model with a valid distance to be able to use the indexing')
        if verbose:
            print('[Indexing %d points]' % len(x))
            print('|-Computing embeddings')
        predictions = self.predict(x)
        self._index.batch_add(predictions,
                              y,
                              data,
                              build=build,
                              verbose=verbose)

    def lookup(self,
               x: Tensor,
               k: int = 5,
               verbose: int = 1) -> List[List[Lookup]]:
        """Find the k closest matches in the index for a set of samples.

        Args:
            x: Samples to match.

            k: Number of nearest neighboors to lookup. Defaults to 5.

            verbose: display progress. Default to 1.

        Returns
            list of list of k nearest neighboors:
            List[List[Lookup]]
        """
        predictions = self.predict(x)
        return self._index.batch_lookup(predictions,
                                        k=k,
                                        verbose=verbose)

    def single_lookup(self,
                      x: Tensor,
                      k: int = 5) -> List[Lookup]:
        """Find the k closest matches in the index for a given sample.

        Args:
            x: Sample to match.

            k: Number of nearest neighboors to lookup. Defaults to 5.

        Returns
            list of the k nearest neigboors info:
            List[Lookup]
        """
        x = tf.expand_dims(x, axis=0)
        prediction = self.predict(x)
        return self._index.single_lookup(prediction, k=k)

    def index_summary(self):
        "Display index info summary."
        self._index.print_stats()

    def calibrate(self,
                  x: FloatTensor,
                  y: IntTensor,
                  thresholds_targets: Dict[str, float] = {},
                  k: int = 1,
                  calibration_metric: Union[str, EvalMetric] = "f1_score",
                  extra_metrics: List[Union[str, EvalMetric]] = ['accuracy', 'recall'],  # noqa
                  rounding: int = 2,
                  verbose: int = 1):
        """Calibrate model thresholds using a test dataset.
            FIXME: more detailed explaination.

            Args:

                x: examples to use for the calibration.

                y: Expected labels for the nearest neighboors.

                thresholds_targets: Dict of performance targets to
                (if possible) meet with respect to the `calibration_metric`.

                calibration_metric: [Metric()](metrics/overview.md) used to
                evaluate the performance of the index.

                k: How many neighboors to use during the calibration.
                Defaults to 1.

                extra_metrics: List of additional [Metric()](
                metrics/overview.md) to compute and report.

                rounding: Metric rounding. Default to 2 digits.

                verbose: Be verbose and display calibration results.
                Defaults to 1.

            Returns:
                Calibration results: `{"cutpoints": {}, "thresholds": {}}`
        """

        # predict
        predictions = self.predict(x)

        # calibrate
        return self._index.calibrate(predictions=predictions,
                                     y=y,
                                     thresholds_targets=thresholds_targets,
                                     k=k,
                                     calibration_metric=calibration_metric,
                                     extra_metrics=extra_metrics,
                                     rounding=rounding,
                                     verbose=verbose)

    def match(self,
              x: FloatTensor,
              cutpoint='optimal',
              no_match_label=-1,
              verbose=0):
        """Match a set of examples against the calibrated index

        For the match function to work, the index must be calibrated using
        calibrate().

        Args:
            x: Batch of examples to be matched against the index.

            cutpoint: Which calibration threshold to use.
            Defaults to 'optimal' which is the optimal F1 threshold computed
            using calibrate().

            no_match_label: Which label value to assign when there is no
            match. Defaults to -1.

            verbose. Be verbose. Defaults to 0.

        Returns:
            List of class ids that matches for each supplied example

        Notes:
            This function matches all the cutpoints at once internally as there
            is little performance downside to do so and allows to do the
            evaluation in a single go.

        """
        # basic checks
        if not self._index.is_calibrated:
            raise ValueError('Uncalibrated model: run model.calibration()')

        # get predictions
        predictions = self.predict(x)

        # matching
        matches = self._index.match(predictions,
                                    no_match_label=no_match_label,
                                    verbose=verbose)

        # select which matches to return
        if cutpoint == 'all':  # returns all the cutpoints for eval purpose.
            return matches
        else:  # normal match behavior - returns a specific cut point
            return matches[cutpoint]

    def evaluate_matching(self,
                          x: Tensor,
                          y: IntTensor,
                          k: int = 1,
                          extra_metrics: List[Union[EvalMetric, str]] = ['accuracy', 'recall'],  # noqa
                          verbose: int = 1):
        """Evaluate model matching accuracy on a given evaluation dataset.

        Args:
            x: Examples to be matched against the index.

            y: Label associated with the examples supplied.

            k: How many neigboors to use to perform the evaluation.
            Defaults to 1.

            extra_metrics: Additional (distance_metrics.mde)[distance metrics]
            to be computed during the evaluation. Defaut to accuracy and
            recall.

            verbose (int, optional): Display results if set to 1 otherwise
            results are returned silently. Defaults to 1.

        Returns:
            Dictionnary of (distance_metrics.md)[evaluation metrics]
        """
        # There is some code duplication in this function but that is the best
        # solution to keep the end-user API clean and doing inferences once.

        if not self._index.is_calibrated:
            raise ValueError('Uncalibrated model: run model.calibration()')

        # get embeddings
        if verbose:
            print("|-Computing embeddings")
        predictions = self.predict(x)

        results: DefaultDict['str', Dict[str, Union[int, float, str]]] = defaultdict(dict)  # noqa
        cal_metric = self._index.get_calibration_metric()

        if verbose:
            pb = tqdm(total=len(self._index.cutpoints),
                      desc='Evaluating cutpoints')

        res: Dict[str, Union[str, int, float]] = {}
        for cp_name, cp_data in self._index.cutpoints.items():
            # create a metric that match at the requested k and threshold
            metric = make_metric(cal_metric.name)
            metric.k = k
            metric.distance_threshold = float(cp_data['distance'])
            metrics = copy(extra_metrics)
            metrics.append(metric)
            res.update(self._index.evaluate(predictions, y, metrics, k))
            res['distance'] = cp_data['distance']
            res['name'] = cp_name
            results[cp_name] = res
            if verbose:
                pb.update()
        if verbose:
            pb.close()

        if verbose:
            headers = ['name', cal_metric.name]
            for i in results['optimal'].keys():
                if i not in headers:
                    headers.append(str(i))
            rows = []
            for data in results.values():
                rows.append([data[v] for v in headers])
            print('\n [Summary]\n')
            print(tabulate(rows, headers=headers))

        return results

    def reset_index(self):
        "Reinitialize the index"
        self._index.reset()

    def index_size(self) -> int:
        "Return the index size"
        return self._index.size()

    def load_index(self, filepath: str):
        """Load Index data from a checkpoint and initialize underlying
        structure with the reloaded data.

        Args:
            path: Directory where the checkpoint is located.
            verbose: Be verbose. Defaults to 1.
        """

        index_path = Path(filepath) / "index"
        self._index = Indexer.load(index_path)

    def save_index(self, filepath, compression=True):
        """Save the index to disk

        Args:
            path: directory where to save the index
            compression: Store index data compressed. Defaults to True.
        """
        index_path = Path(filepath) / "index"
        self._index.save(index_path, compression=compression)

    def save(self,
             filepath: str,
             save_index: bool = True,
             compression: bool = True,
             overwrite: bool = True,
             include_optimizer: bool = True,
             signatures=None,
             options=None,
             save_traces: bool = True):
        """Save the model and the index.

        Args:
            filepath: where to save the model.

            save_index: Save the index content. Defaults to True.

            compression: Compress index data. Defaults to True.

            overwrite: Overwrite previous model. Defaults to True.

            include_optimizer: Save optimizer state. Defaults to True.

            signatures: Signatures to save with the model. Defaults to None.

            options: A `tf.saved_model.SaveOptions` to save with the model.
            Defaults to None.

            save_traces (optional): When enabled, the SavedModel will
            store the function traces for each layer. This can be disabled,
            so that only the configs of each layer are stored.
            Defaults to True. Disabling this will decrease serialization
            time and reduce file size, but it requires that all
            custom layers/models implement a get_config() method.
        """

        # save trace doesn't exist prior to 2.4 -- asking for it but not
        # using it

        # call underlying keras method to save the mode graph and its weights
        tf.keras.models.save_model(self,
                                   filepath,
                                   overwrite=overwrite,
                                   include_optimizer=include_optimizer,
                                   signatures=signatures,
                                   options=options,
                                   save_traces=save_traces)
        if hasattr(self, '_index') and self._index and save_index:
            self.save_index(filepath, compression=compression)
        else:
            print('Index not saved as save_index=False')

    def to_data_frame(self, num_items: int = 0) -> PandasDataFrame:
        """Export data as pandas dataframe

        Args:
            num_items (int, optional): Num items to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """
        return self._index.to_data_frame(num_items=num_items)

    # We don't need from_config as the index is reloaded separatly.
    # this is kept as a reminder that it was looked into and decided to split
    # the index reloading instead of overloading this method.
    # @classmethod
    # def from_config(cls, config):
    #     return super().from_config(**config)



"EfficientNet backbone for similarity learning"
import re
from typing import Tuple, Callable, Union
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.applications import efficientnet
from tensorflow_similarity.models import SimilarityModel

EFF_INPUT_SIZE = {
    "B0": 224,
    "B1": 240,
    "B2": 260,
    "B3": 300,
    "B4": 380,
    "B5": 456,
    "B6": 528,
    "B7": 600,
}

EFF_ARCHITECTURE = {
    "B0": efficientnet.EfficientNetB0,
    "B1": efficientnet.EfficientNetB1,
    "B2": efficientnet.EfficientNetB2,
    "B3": efficientnet.EfficientNetB3,
    "B4": efficientnet.EfficientNetB4,
    "B5": efficientnet.EfficientNetB5,
    "B6": efficientnet.EfficientNetB6,
    "B7": efficientnet.EfficientNetB7,
}


# Create an image augmentation pipeline.
def EfficientNetSim(input_shape: Tuple[int],
                    embedding_size: int = 128,
                    variant: str = "B0",
                    weights: str = "imagenet",
                    augmentation: Union[Callable, str] = "basic",
                    trainable: str = "frozen"):
    """Build an EffecientNet Model backbone for similarity learning

    Architecture from [EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks
](https://arxiv.org/abs/1905.11946)

    Args:
        input_shape: Size of the image input prior to augmentation,
        must be bigger than the size of Effnet version you use. See below for
        min input size.

        embedding_size: Size of the output embedding. Usually between 64
        and 512. Defaults to 128.
        variant: Which Variant of the EfficientNEt to use. Defaults to "B0".

        weights: Use pre-trained weights - the only available currently being
        imagenet. Defaults to "imagenet".

        augmentation: How to augment the data - either pass a Sequential model
        of keras.preprocessing.layers or use the built in one or set it to
        None to disable. Defaults to "basic".

        trainable: Make the EfficienNet backbone fully trainable or partially
        trainable. Either "full" to make the entire backbone trainable,
        "partial" to only make the last 3 block trainable or "frozen" to make
        it not trainable. Defaults to "frozen".

    Note:
        EfficientNet expects images at the following size:
         - "B0": 224,
         - "B1": 240,
         - "B2": 260,
         - "B3": 300,
         - "B4": 380,
         - "B5": 456,
         - "B6": 528,
         - "B7": 600,

    """

    # input
    inputs = layers.Input(shape=input_shape)
    x = inputs

    if variant not in EFF_INPUT_SIZE:
        raise ValueError("Unknown efficientnet variant. Valid B0...B7")
    img_size = EFF_INPUT_SIZE[variant]

    # augmentation
    # add the basic version or the suppplied one.
    if callable(augmentation):
        x = augmentation(x)
    elif augmentation == "basic":
        # augs usually used in benchmark and work almost always well
        over_size = int(img_size * 1.5)
        x = tf.keras.Sequential([
            layers.experimental.preprocessing.Resizing(over_size, over_size),
            layers.experimental.preprocessing.RandomCrop(img_size, img_size),
            layers.experimental.preprocessing.RandomFlip("horizontal")
        ])(x)
    else:
        print(f'Warning: {augmentation} is not supported and will be skipped.')

    x = build_effnet(x, variant, weights, trainable)
    x = layers.GlobalAveragePooling2D()(x)
    outputs = layers.Dense(embedding_size)(x)
    return SimilarityModel(inputs, outputs)


def build_effnet(x, variant, weights, trainable):
    "Build the requested efficient net."

    # init
    effnet_fn = EFF_ARCHITECTURE[variant]
    effnet = effnet_fn(weights=weights, include_top=False)

    if trainable == "full":
        effnet.trainable = True
    elif trainable == "partial":
        # let's mark the top part of the network as trainable
        effnet.trainable = True
        for layer in effnet.layers:
            # Freeze all the layers before the the last 3 blocks
            if not re.search("^block[5,6,7]|^top", layer.name):
                layer.trainable = False
            # don't change the batchnorm weights
            if isinstance(layer, layers.BatchNormalization):
                layer.trainable = False
    else:
        effnet.trainable = False

    # wire
    x = efficientnet.preprocess_input(x)
    x = effnet(x)

    return x



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from setuptools import find_packages
from setuptools import setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


long_description = "FIXME"

setup(
    name="tensorflow_similarity",
    version=get_version("tensorflow_similarity/__init__.py"),
    description="Metric Learning for Humans",
    long_description=long_description,
    author='Tensorflow Similarity authors',
    author_email='tf-similarity@google.com',
    url='https://github.com/tensorflow/similarity',
    license='Apache License 2.0',
    install_requires=[
        'numpy', 'tabulate', 'nmslib', 'tensorflow>=2.4', 'tqdm',
        'matplotlib', 'pandas', 'tensorflow-datasets>=4.2'
    ],
    extras_require={
        "dev": [
            'setuptools', 'wheel', 'mypy', 'pytest', 'flake8', 'pytest-cov',
            'pytype', 'twine', 'setuptools', 'nmslib', 'tensorflow', 'mkdocs',
            'mkdocs-material', 'mkdocs-autorefs', 'mkdocstrings'
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    packages=find_packages())



import numpy as np
import tensorflow as tf
from tensorflow_similarity.models import SimilarityModel
from tensorflow_similarity.callbacks import EvalCallback
from tensorflow_similarity.callbacks import SplitValidationLoss


def test_eval_callback(tmp_path):
    queries = tf.constant([[1, 2], [1, 2]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [1, 2]])
    targets_labels = tf.constant([1, 2])

    log_dir = tmp_path / 'sec/'
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(queries,
                            query_labels,
                            targets,
                            targets_labels,
                            tb_logdir=str(log_dir))
    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2, ))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile('adam', loss='mse', distance='cosine')
    callback.model = model

    # call the only callback method implemented
    callback.on_epoch_end(0, {})


def test_split_val_loss_callback(tmp_path):
    x = tf.constant([[-1]]*5+[[1]]*5)
    y = tf.constant([0]*5+[1]*5)
    known_classes = np.array([1])

    callback = SplitValidationLoss(x, y, known_classes)

    class MockModel:
        def evaluate(self, x, y, verbose=0):
            _, _ = x, verbose
            return float(tf.math.reduce_sum(y) / tf.shape(y)[0])

    callback.model = MockModel()

    # call the only callback method implemented
    logs = {}
    callback.on_epoch_end(0, logs=logs)

    assert logs['known_val_loss'] == 1.0
    assert logs['unknown_val_loss'] == 0.0



import tensorflow as tf


def is_tensor_or_variable(x):
    "check if a variable is tf.Tensor or tf.Variable"
    return tf.is_tensor(x) or isinstance(x, tf.Variable)


def tf_cap_memory():
    "Avoid TF to hog memory before needing it"
    gpus = tf.config.experimental.list_physical_devices('GPU')

    if gpus:
        for gpu in gpus:
            try:
                tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)



"Index embedding to allow distance based lookup"

import json
from time import time
import numpy as np
from collections import defaultdict, deque
from tabulate import tabulate
from pathlib import Path
import tensorflow as tf
from tqdm.auto import tqdm

# types
from typing import Dict, List, Union, DefaultDict, Deque, Optional
from .types import FloatTensor, Lookup, PandasDataFrame, Tensor

# internal
from .distances import distance_canonicalizer, Distance
from .matchers import Matcher, NMSLibMatcher
from .tables import Table, MemoryTable
from .evaluators import Evaluator, MemoryEvaluator
from .metrics import EvalMetric, make_metric, F1Score


class Indexer():
    """Indexing system that allows to efficiently find nearest embeddings
    by indexing known embeddings and make them searchable using an
    [Approximate Nearest Neigboors Search](https://en.wikipedia.org/wiki/Nearest_neighbor_search)
    search implemented via the [`Matcher()`](matchers/overview.md) classes
    and associated data lookup via the [`Table()`](tables/overview.md) classes.

    The indexer allows to evaluate the quality of the constructed index and
    calibrate the [SimilarityModel.match()](similarity_model.md) function via
    the [`Evaluator()`](evaluators/overview.md) classes.
    """
    # semantic sugar for the order of the returned data
    EMBEDDINGS = 0
    DISTANCES = 1
    LABELS = 2
    DATA = 3
    RANKS = 4

    def __init__(self,
                 embedding_size: int,
                 distance: Union[Distance, str] = 'cosine',
                 matcher: Union[Matcher, str] = 'nmslib',
                 table: Union[Table, str] = 'memory',
                 evaluator: Union[Evaluator, str] = 'memory',
                 embedding_output: int = None,
                 stat_buffer_size: int = 1000) -> None:
        """Index embeddings to make them searchable via KNN

        Args:
            embedding_size: Size of the embeddings that will be stored.
            It is usually equivalent to the size of the output layer.

            distance: Distance used to compute embeddings proximity.
            Defaults to 'cosine'.

            table: How to store the index records.
            Defaults to 'memory'.

            matcher: Which `Matcher()` framework to use to perfom KNN
            search. Defaults to 'nmslib'.

            evaluator: What type of `Evaluator()` to use to evaluate index
            performance. Defaults to in-memory one.

            embedding_output: Which model output head predicts
            the embbedings that should be indexed. Default to None which is for
            single output model. For multi-head model, the callee, usually the
            `SimilarityModel()` class is responsible for passing the
            correct one.

            stat_buffer_size: Size of the sliding windows
            buffer used to computer index performance. Defaults to 1000.

        Raises:
            ValueError: Invalid matcher or table.
        """
        distance = distance_canonicalizer(distance)
        self.distance = distance  # needed for save()/load()
        self.embedding_output = embedding_output
        self.embedding_size = embedding_size

        # internal structure naming
        # FIXME support custom objects
        self.matcher_type = matcher
        self.table_type = table
        self.evaluator_type = evaluator

        # stats configuration
        self.stat_buffer_size = stat_buffer_size

        # calibration
        self.is_calibrated = False
        self.calibration_metric: EvalMetric = F1Score()
        self.cutpoints: Dict[str, Dict[str, Union[int, float, str]]] = {}
        self.calibration_thresholds: Dict[str, List[Union[float, int]]] = {}

        # initialize internal structures
        self._init_structures()

    def reset(self) -> None:
        "Reinitialize the indexer"
        self._init_structures()

    def _init_structures(self) -> None:
        "(re)intialize internal storage structure"

        if self.matcher_type == 'nmslib':
            self.matcher: Matcher = NMSLibMatcher(distance=self.distance,
                                                  dims=self.embedding_size)
        elif isinstance(self.matcher_type, Matcher):
            self.matcher = self.matcher_type
        else:
            raise ValueError("You need to either supply a known matcher name\
                or a Matcher() object")

        # mapper from id to record data
        if self.table_type == 'memory':
            self.table: Table = MemoryTable()
        elif isinstance(self.table_type, Table):
            self.table = self.table_type
        else:
            raise ValueError("You need to either supply a know table name\
                or a Table() object")

        # code used to evaluate indexer performance
        if self.evaluator_type == 'memory':
            self.evaluator: Evaluator = MemoryEvaluator()
        elif isinstance(self.evaluator_type, Evaluator):
            self.evaluator = self.evaluator_type
        else:
            raise ValueError("You need to either supply a know evaluator name\
                or an Evaluator() object")
        # stats
        self._stats: DefaultDict[str, int] = defaultdict(int)
        self._lookup_timings_buffer: Deque = deque([], maxlen=self.stat_buffer_size)  # noqa

        # calibration data
        self.is_calibrated = False
        self.calibration_metric = F1Score()
        self.cutpoints = {}
        self.calibration_thresholds = {}

    def _get_embedding(self, prediction: FloatTensor) -> FloatTensor:
        """Return the 1st embedding vector from a (multi-output) model
        prediction

        See: `single_lookup()`, `add()`

        Args:
            prediction: Model prediction for single embedding.

        Returns:
            FloatTensor: 1D Tensor that contains the actual embedding
        """

        if isinstance(self.embedding_output, int):
            # in multi-output: embedding is [output_num][0]
            embedding: FloatTensor = prediction[self.embedding_output][0]
        else:
            # single output > return 1st element
            embedding = prediction[0]
        return embedding

    def _get_embeddings(self, predictions: FloatTensor) -> FloatTensor:
        """Return the embedding vectors from a (multi-output) model prediction

        Args:
            prediction: Model predictions.

        Returns:
            Tensor: 2D Tensor (num_embeddings, embedding_value)
        """

        if isinstance(self.embedding_output, int):
            embeddings: FloatTensor = predictions[self.embedding_output]
        else:
            # needed for typing
            embeddings = predictions
        return embeddings

    def _cast_label(self, label: Optional[int]) -> Optional[int]:
        if label is not None:
            label = int(label)
        return label

    def add(self,
            prediction: FloatTensor,
            label: Optional[int] = None,
            data: Tensor = None,
            build: bool = True,
            verbose: int = 1):
        """ Add a single embedding to the indexer

        Args:
            prediction: TF similarity model prediction.

            label: Label(s) associated with the
            embedding. Defaults to None.

            data: Input data associated with
            the embedding. Defaults to None.

            build: Rebuild the index after insertion.
            Defaults to True. Set it to false if you would like to add
            multiples batchs/points and build it manually once after.

            verbose: Display progress if set to 1.
            Defaults to 1.
        """

        # deal with potential multi-output
        embedding = self._get_embedding(prediction)

        # store data and get its id
        idx = self.table.add(embedding, label, data)

        # add index to the embedding
        self.matcher.add(embedding, idx, build=build, verbose=verbose)

    def batch_add(self,
                  predictions: FloatTensor,
                  labels: Optional[List[Optional[int]]] = None,
                  data: Optional[Tensor] = None,
                  build: bool = True,
                  verbose: int = 1):
        """Add a batch of embeddings to the indexer

        Args:
            predictions: TF similarity model predictions.

            labels: label(s) associated with the embedding. Defaults to None.

            datas: input data associated with the embedding. Defaults to None.

            build: Rebuild the index after insertion.
            Defaults to True. Set it to false if you would like to add
            multiples batchs/points and build it manually once after.

            verbose: Display progress if set to 1. Defaults to 1.
        """

        # deal with potential multi-output
        embeddings = self._get_embeddings(predictions)

        # store points
        if verbose:
            print('|-Storing data points in index table')
        idxs = self.table.batch_add(embeddings, labels, data)
        self.matcher.batch_add(embeddings, idxs, build=build, verbose=verbose)

    def single_lookup(
            self,
            prediction: FloatTensor,
            k: int = 5) -> List[Lookup]:
        """Find the k closest matches of a given embedding

        Args:
            prediction: model prediction.
            k: Number of nearest neighboors to lookup. Defaults to 5.
        Returns
            list of the k nearest neigboors info:
            List[Lookup]
        """

        embedding = self._get_embedding(prediction)
        start = time()
        idxs, distances = self.matcher.lookup(embedding, k=k)
        nn_embeddings, labels, data = self.table.batch_get(idxs)

        lookup_time = time() - start
        lookups = []
        for i in range(len(nn_embeddings)):
            # ! casting is needed to avoid slowness down the line
            lookups.append(Lookup(
                rank=i + 1,
                embedding=nn_embeddings[i],
                distance=float(distances[i]),
                label=self._cast_label(labels[i]),
                data=data[i]
                ))
        self._lookup_timings_buffer.append(lookup_time)
        self._stats['num_lookups'] += 1
        return lookups

    def batch_lookup(self,
                     predictions: FloatTensor,
                     k: int = 5,
                     verbose: int = 1) -> List[List[Lookup]]:

        """Find the k closest matches for a set of embeddings

        Args:
            predictions: model predictions.
            k: Number of nearest neighboors to lookup. Defaults to 5.
            verbose: Be verbose. Defaults to 1.

        Returns
            list of list of k nearest neighboors:
            List[List[Lookup]]
        """

        embeddings = self._get_embeddings(predictions)
        num_embeddings = len(embeddings)
        start = time()
        batch_lookups = []

        if verbose:
            print("\nPerforming NN search\n")
        batch_idxs, batch_distances = self.matcher.batch_lookup(embeddings,
                                                                k=k)

        if verbose:
            pb = tqdm(total=num_embeddings, desc='Building NN list')
        for eidx in range(num_embeddings):
            lidxs = batch_idxs[eidx]   # list of nn idxs
            distances = batch_distances[eidx]

            nn_embeddings, labels, data = self.table.batch_get(lidxs)
            lookups = []
            for i in range(len(nn_embeddings)):
                # ! casting is needed to avoid slowness down the line
                lookups.append(Lookup(
                    rank=i + 1,
                    embedding=nn_embeddings[i],
                    distance=float(distances[i]),
                    label=self._cast_label(labels[i]),
                    data=data[i]
                ))
            batch_lookups.append(lookups)

            if verbose:
                pb.update()

        if verbose:
            pb.close()

        # stats
        lookup_time = time() - start
        per_lookup_time = lookup_time / num_embeddings
        for _ in range(num_embeddings):
            self._lookup_timings_buffer.append(per_lookup_time)
        self._stats['num_lookups'] += num_embeddings

        return batch_lookups

    # evaluation related functions
    def evaluate(self,
                 predictions: FloatTensor,
                 y: List[int],
                 metrics: List[Union[str, EvalMetric]],
                 k: int = 1,
                 verbose: int = 1) -> Dict[str, Union[float, int]]:
        """Evaluate the quality of the index against a test dataset.

        Args:
            predictions: Test emebddings computed by the SimilarityModel.
            y: Expected labels for the nearest neighboors.
            metrics: List of [Metric()](metrics/overview.md) to compute.
            k: How many neighboors to use during the evaluation. Defaults to 1.
            verbose: Be verbose. Defaults to 1.

        Returns:
            Dictionary of metric results where keys are the metric names and
            values are the metrics values.
        """
        # Find NN
        lookups = self.batch_lookup(predictions, verbose=verbose)

        # Evaluate them
        return self.evaluator.evaluate(index_size=self.size(),
                                       metrics=metrics,
                                       targets_labels=y,
                                       lookups=lookups)

    def calibrate(self,
                  predictions: FloatTensor,
                  y: List[int],
                  thresholds_targets: Dict[str, float],
                  calibration_metric: Union[str, EvalMetric] = "f1_score",
                  k: int = 1,
                  extra_metrics: List[Union[str, EvalMetric]] = ['accuracy', 'recall'],  # noqa
                  rounding: int = 2,
                  verbose: int = 1
                  ) -> Dict[str, Union[Dict[str, float], List[float]]]:
        """Calibrate model thresholds using a test dataset.

        FIXME: more detailed explaination.

        Args:

            predictions: Test emebddings computed by the SimilarityModel.

            y: Expected labels for the nearest neighboors.

            thresholds_targets: Dict of performance targets to (if possible)
            meet with respect to the `calibration_metric`.

            calibration_metric: [Metric()](metrics/overview.md) used to
            evaluate the performance of the index.

            k: How many neighboors to use during the calibration.
            Defaults to 1.

            extra_metrics: List of additional [Metric()](metrics/overview.md)
            to compute and report.

            rounding: Metric rounding. Default to 2 digits.

            verbose: Be verbose and display calibration results. Defaults to 1.

        Returns:
           Calibration results: `{"cutpoints": {}, "thresholds": {}}`
        """

        # find NN
        lookups = self.batch_lookup(predictions, verbose=verbose)

        # making sure our metrics are all EvalMetric object
        calibration_metric = make_metric(calibration_metric)
        # This aweful syntax is due to mypy not understanding subtype :(
        extra_eval_metrics: List[Union[str, EvalMetric]] = [make_metric(m) for m in extra_metrics]  # noqa
        # running calibration
        thresholds, cutpoints = self.evaluator.calibrate(
            self.size(),
            calibration_metric=calibration_metric,
            thresholds_targets=thresholds_targets,
            targets_labels=y,
            lookups=lookups,
            extra_metrics=extra_eval_metrics,
            metric_rounding=rounding,
            verbose=verbose
        )

        # display cutpoint results if requested
        if verbose:
            headers = ['name', 'value', 'distance']  # noqa

            # dynamicaly find which metrics we need
            for data in cutpoints.values():
                for k in data.keys():
                    if k not in headers:
                        headers.append(str(k))
                break

            # print(cutpoints)
            rows = []
            for data in cutpoints.values():
                rows.append([data[v] for v in headers])
            print("\n", tabulate(rows, headers=headers))

        # store info for serialization purpose
        self.is_calibrated = True
        self.calibration_metric = calibration_metric
        self.cutpoints = cutpoints
        self.calibration_thresholds = thresholds
        return {"cutpoints": cutpoints, "thresholds": thresholds}

    def match(self,
              predictions: FloatTensor,
              no_match_label: int = -1,
              verbose: int = 1) -> Dict[str, List[int]]:
        """Match embeddings against the various cutpoints thresholds

        Args:
            predictions (FloatTensor): embeddings

            no_match_label (int, optional): What label value to assign when
            there is no match. Defaults to -1.

            verbose (int): display progression. Default to 1.

        Notes:

            1. It is up to the [`SimilarityModel.match()`](similarity_model.md)
            code to decide which of cutpoints results to use / show to the
            users. This function returns all of them as there is little
            performance downside to do so and it makes the code clearer
            and simpler.

            2. The calling function is responsible to return the list of class
            matched to allows implementation to use additional criterias
            if they choose to.

        Returns:
            Dict of matches list keyed by cutpoint names.
        """
        lookups = self.batch_lookup(predictions, k=1, verbose=verbose)

        # vectorize
        distances = []
        labels = []
        for lookup in lookups:
            distances.append(lookup[0].distance)
            labels.append(lookup[0].label)

        if verbose:
            pb = tqdm(total=len(distances) * len(self.cutpoints),
                      desc='matching embeddings')

        matches: Dict = defaultdict(list)
        for name, cp in self.cutpoints.items():
            threshold = float(cp['distance'])
            for idx, distance in enumerate(distances):
                if distance <= threshold:
                    label = labels[idx]
                else:
                    label = no_match_label
                matches[name].append(label)

                if verbose:
                    pb.update()

        if verbose:
            pb.close()

        return matches

    def save(self, path: str, compression: bool = True):
        """Save the index to disk

        Args:
            path: directory where to save the index
            compression: Store index data compressed. Defaults to True.
        """
        path = str(path)
        # saving metadata
        metadata = {
            "size": self.size(),
            "compression": compression,
            "distance": self.distance.name,
            "embedding_output": self.embedding_output,
            "embedding_size": self.embedding_size,

            "table": self.table_type,
            "evaluator": self.evaluator_type,
            "matcher": self.matcher_type,

            "stat_buffer_size": self.stat_buffer_size,
            "is_calibrated": self.is_calibrated,

            "calibration_metric_config": self.calibration_metric.get_config(),
            "cutpoints": self.cutpoints,
            "calibration_thresholds": self.calibration_thresholds
        }

        metadata_fname = self.__make_metadata_fname(path)
        tf.io.write_file(metadata_fname, json.dumps(metadata))

        self.table.save(path, compression=compression)
        self.matcher.save(path)

    @staticmethod
    def load(path: Union[str, Path], verbose: int = 1):
        """Load Index data from a checkpoint and initialize underlying
        structure with the reloaded data.

        Args:
            path: Directory where the checkpoint is located.
            verbose: Be verbose. Defaults to 1.

        Returns:
            Initialized index
        """
        path = str(path)
        # recreate the index from metadata
        metadata_fname = Indexer.__make_metadata_fname(path)
        metadata = tf.io.read_file(metadata_fname)
        metadata = tf.keras.backend.eval(metadata)
        md = json.loads(metadata)
        index = Indexer(distance=md['distance'],
                        embedding_size=md['embedding_size'],
                        embedding_output=md['embedding_output'],
                        table=md['table'],
                        evaluator=md['evaluator'],
                        matcher=md['matcher'],
                        stat_buffer_size=md['stat_buffer_size'])

        # reload the tables
        if verbose:
            print("Loading index data")
        index.table.load(path)

        # rebuild the index
        if verbose:
            print('Loading index matcher')
        index.matcher.load(path)

        # reload calibration data if any
        index.is_calibrated = md['is_calibrated']
        if index.is_calibrated:
            if verbose:
                print("Loading calibration data")
            index.calibration_metric = EvalMetric.from_config(md['calibration_metric_config'])  # noqa
            index.cutpoints = md['cutpoints']
            index.calibration_thresholds = md['calibration_thresholds']
        return index

    def get_calibration_metric(self):
        return self.calibration_metric

    def size(self) -> int:
        "Return the index size"
        return self.table.size()

    def stats(self):
        """return index statistics"""
        stats = self._stats
        stats['size'] = self.table.size()
        stats['stat_buffer_size'] = self.stat_buffer_size

        # query performance - make sure we don't count unused buffer
        max_idx = min(stats['num_lookups'], self.stat_buffer_size)
        lookup_timings = list(self._lookup_timings_buffer)[:max_idx]

        # ensure we never have an empty list
        lookup_timings = lookup_timings if list(lookup_timings) else [0]

        # compute stats
        stats['query_performance'] = {
            'min': np.min(lookup_timings),
            'max': np.max(lookup_timings),
            'avg': np.average(lookup_timings),
            'median': np.median(lookup_timings),
            'stddev': np.std(lookup_timings)
        }
        return stats

    def print_stats(self):
        "display statistics in terminal friendly fashion"
        # compute statistics
        stats = self.stats()

        # info
        print('[Info]')
        rows = [
            ['distance', self.distance],
            ['index table', self.table_type],
            ['matching algorithm', self.matcher_type],
            ['evaluator', self.evaluator_type],
            ['index size', self.size()],
            ['calibrated', self.is_calibrated],
            ['calibration_metric', self.calibration_metric],
            ['embedding_output', self.embedding_output]
        ]
        print(tabulate(rows))
        print('\n')

        print('\n[Performance]')
        rows = [['num lookups', stats['num_lookups']]]
        for k, v in stats['query_performance'].items():
            rows.append([k, v])
        print(tabulate(rows))

    def to_data_frame(self, num_items: int = 0) -> PandasDataFrame:
        """Export data as pandas dataframe

        Args:
            num_items (int, optional): Num items to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """
        return self.table.to_data_frame(num_items)

    @staticmethod
    def __make_metadata_fname(path):
        return str(Path(path) / 'index_metadata.json')
